from OritatamiSimGrid import *
from OritatamiSimNodes import *
from OritatamiSimUtils import *
from SimulatorFramework import SimCanvas
from SimulatorFramework import SimUtils
import wx
import thread
import threading
import os
import sys
import random
import math
from wx.lib.pubsub import pub
import svgwrite
from svgwrite import cm, mm


SHOW_GRID_AND_COORDS = 0
SHOW_GRID_ONLY       = 1
SHOW_NO_GRID         = 2


ORIT_MODE_PAN_ZOOM     = 0
ORIT_MODE_DRAW_BEAD    = 1
ORIT_MODE_DRAW_LINK    = 2
ORIT_MODE_DRAW_BOND    = 3

SIMFW_FULLDISPLAY      = 0
#SIMFW_GROWTHDISPLAY    = 1
SIMFW_NODISPLAY        = 1


class OritatamiSimCanvas(SimCanvas.Canvas):
    def __init__(self, parent, idIn, frame, size, app):

        self.x = self.y = 0
        self.seed = {}
        self.currentBreakpoint = -1
        self.breakOnNondeterminism = False
        self.confirmOverArity = True
        self.confirmInvalidBonds = True
        
        self.showGrid = SHOW_GRID_AND_COORDS
        
        self.mouseMode = ORIT_MODE_PAN_ZOOM
        self.drawingModeBeadType = None

        self.pen_cache = {}
        self.brush_cache = {}
        self.gridPen = self.ColoredPen("BLACK", 1)
        self.axesPen = self.ColoredPen("RED", 1)
        self.bondPen = self.ColoredPen("RED", 2)
        self.linkPen = self.ColoredPen("BLACK", 4)
        self.grabbingCursorImg = wx.Image("images/grabbing.gif", wx.BITMAP_TYPE_GIF)

        self.sideLen = TRISIZE
        self.triangleHeight = math.tan(math.radians(60)) * 0.5 * self.sideLen

        #self.gridWidth = GRIDSTARTSIZE
        #self.gridHeight = GRIDSTARTSIZE
        
        self.dragStart   = None
        self.linkStart   = None
        self.linkEnd     = None
        self.posSelected = None
        self.negSelected = None
        self.lastCurrent = None

        SimCanvas.Canvas.__init__(self, parent, idIn, frame, size, app)
        self.SetScrollRate(1,1)
		
    def CreateNewAssembly(self):
        self.assembly = ONodeGrid(GRIDSTARTSIZE, GRIDSTARTSIZE)
        self.gridWidth = self.assembly.GetWidth()
        self.gridHeight = self.assembly.GetHeight()
        self.drawnAssembly = []
        self.DrawQueue = []

    def SendStatusBarUpdate(self, field, msg, bForce=True):
        if bForce == True or self.keepGoing == False:
            wx.CallAfter(pub.getDefaultPublisher().sendMessage, "updateStatusbar", msg = (field, msg))

    def DoLoad(self, bAssembly=True, bSystem=False):
        typeString = "Oritatami assembly files (*.oass)|*.oass"
        if bAssembly == False:
            typeString = "Oritatami system files (*.orit)|*.orit"
        elif bAssembly == True and bSystem == True:
            typeString = "Oritatami system/seed files (*.orit;*.oass)|*.orit;*.oass"
        openFileDialog = wx.FileDialog(self, "Open", "", "", 
                                       typeString,
                                       wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
        openFileDialog.ShowModal()
        fileName = openFileDialog.GetPath()
        openFileDialog.Destroy()
        if (len(fileName) > 0):
            
            dlg = wx.TextEntryDialog(self, "You can set your random seed below\n(use -1 to choose a new random value):", 
                                     caption="Set random number seed", value=str(self.randSeed))
            if dlg.ShowModal() == wx.ID_OK:
                try:
                    self.randSeed = int(dlg.GetValue())
                except ValueError:
                    SimUtils.globalApp.ShowErrorDialog("Random seed value must be an integer.")
                    dlg.Destroy()
                    return
            dlg.Destroy()

            self.SetRandomSeed(self.randSeed)

            self.CreateNewAssembly()
            if fileName.endswith(".orit"):
                self.assembly.LoadOritatamiSystem(fileName)
            elif fileName.endswith(".oass"):
                self.assembly.LoadOritatamiAssembly(fileName, True)

                #if len(self.assembly.ONodeTypes) > 0:
                    # Randomly set the current drawing bead type to one of the bead types
                    #self.drawingModeBeadType = self.assembly.ONodeTypes[self.assembly.ONodeTypes.keys()[0]]
                #else:
            self.drawingModeBeadType = None
            #self.InitializeDrawing(self.pdc)
            self.app.AddBeadTypeWindow()
            self.app.DrawBeadCycle()
            self.Refresh()
            self.ScrollToCoords((0,0))

    def DoSaveSystemDef(self, bSaveAs=False):
        fileName = SimUtils.globalApp.systemDefinitionFileFullPath
        if fileName is None or bSaveAs == True:
            saveFileDialog = wx.FileDialog(self, "Save System Definition As", "", "", 
                                           "Oritatami system files (*.orit)|*.orit",
                                           wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
            if not saveFileDialog.ShowModal() == wx.ID_OK:
                SimUtils.globalApp.ShowErrorDialog("You must provide a file name to save the oritatami system definition to.")
                return False
            fileName = saveFileDialog.GetPath()
            SimUtils.globalApp.systemDefinitionFile = fileName
            saveFileDialog.Destroy()

        if (fileName is None) or len(fileName) < 1:
            return False
        
        if not fileName[-5:] == ".orit":
            SimUtils.globalApp.ShowErrorDialog("\nFile name must end with the extension '.orit'")
            return False

        SimUtils.globalApp.systemDefinitionFileFullPath = fileName
        if os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        try:
            btStr = "["
            for bt in self.assembly.ONodeTypes:
                if len(btStr) > 1:
                    btStr = btStr + ", "
                btStr = btStr + self.assembly.ONodeTypes[bt].Output()
            btStr = btStr + "]"
            btLine = "BeadTypes=" + btStr
            print("btLine: " + btLine)
            bcLine = "BeadCycle=" + str(self.assembly.beadCycle)
            print("bcLine: " + bcLine)
            dlLine = "DelayLength=" + str(self.assembly.delayLength)
            print("dlLine: " + dlLine)
            arLine = "Arity=" + str(self.assembly.arity)
            print("arLine:" + arLine)
            lwbLine = "LockWithoutBinding=" + str(self.assembly.lockWithoutBinding)
            print("lwbLine: " + lwbLine)
            bffsLine = "BindFullFloppySegment=" + str(self.assembly.bindFullFloppySegment)
            print("bffsLine:" + bffsLine)

            wFile = open(fileName, "w")
            wFile.write(btLine + "\n")
            wFile.write(bcLine + "\n")
            wFile.write(dlLine + "\n")
            wFile.write(arLine + "\n")
            wFile.write(lwbLine + "\n")
            wFile.write(bffsLine + "\n")
            wFile.close()
            SimUtils.globalApp.OutputMessage("Saved oritatami system definition to file '" + fileName + "'")
            return True
        except:
            SimUtils.globalApp.ShowErrorDialog("\nERROR: Unable to write oritatami system definition file: '" + fileName + "'\n")
            return False

    def DoSetSeed(self):
        dlg = wx.MessageDialog(self, "Are you sure you want to set the current assembly as the seed assembly?", caption="Confirm seed setting")
        if dlg.ShowModal() == wx.ID_OK:
            dlg.Destroy()
            assemblyList = self.GetAssemblyList()
            if len(assemblyList) == 0:
                SimUtils.globalApp.ShowErrorDialog("The seed assembly is empty.  It must contain at least one bead.")
                return False
            self.assembly.seed = assemblyList
            self.assembly.lastSeedLoc = self.assembly.lastBeadLoc
        else:
            dlg.Destroy()

    #  Format for assembly file:
    #  [[<bead location>], [<bead location>], ...]
    #  Format for bead location:
    #  ["<type name>", (<coords>), (<parent coords>), [<list of coords bonded to>]]
    def DoSaveAssembly(self, bOnlySeed=False, bSaveAs=False):
        aStr = "assembly"
        AStr = "Assembly"
        fileName = SimUtils.globalApp.assemblyDefinitionFile
        if bOnlySeed == True:
            aStr = "seed"
            AStr = "Seed"
            fileName = SimUtils.globalApp.seedDefinitionFile
        if fileName is None or bSaveAs == True:
            saveFileDialog = wx.FileDialog(self, "Save " + AStr + " As", "", "", 
                                           "Oritatami " + aStr + " files (*.oass)|*.oass",
                                           wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
            if not saveFileDialog.ShowModal() == wx.ID_OK:
                return False
            fileName = saveFileDialog.GetPath()
            saveFileDialog.Destroy()

        if (fileName is None) or (len(fileName) < 1):
                return False

        if not fileName[-5:] == ".oass":
            SimUtils.globalApp.ShowErrorDialog("\nFile name must end with the extension '.oass'")
            return False

        if os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        if SimUtils.globalApp.systemDefinitionFile is None or len(SimUtils.globalApp.systemDefinitionFile) < 1:
            SimUtils.globalApp.ShowErrorDialog("\nERROR: In order to save an assembly, you must first have a valid oritatami system definition file associated with it. Please save your system definition first.\n")
            return False

        defFile = SimUtils.globalApp.systemDefinitionFile
        if os.path.isabs(defFile) == True:
            oassDir = os.path.dirname(fileName)#os.path.realpath(fileName))
            defFile = os.path.relpath(defFile, oassDir)

        systemDefString = "OritatamiSystemDef=" + defFile
        assemblyString = "Assembly="
        if bOnlySeed == True:
            lastBeadString = "LastBead=" + str(self.assembly.lastSeedLoc)
        else:
            lastBeadString = "LastBead=" + str(self.assembly.lastBeadLoc)
        assemblyList = self.GetAssemblyList(bOnlySeed)
        if len(assemblyList) == 0:
            SimUtils.globalApp.ShowErrorDialog("The seed assembly is empty.  It must contain at least one bead.")
            return False
        assemblyString = assemblyString + str(assemblyList)

        try:
            wFile = open(fileName, "w")
            wFile.write(systemDefString + "\n")
            wFile.write(assemblyString + "\n")
            wFile.write(lastBeadString)
            wFile.close()
            if bOnlySeed == True:
                SimUtils.globalApp.seedDefinitionFile = fileName
            else:
                SimUtils.globalApp.assemblyDefinitionFile = fileName
            SimUtils.globalApp.OutputMessage("Saved oritatami " + aStr + " to file '" + fileName + "'")
            return True
        except:
            SimUtils.globalApp.ShowErrorDialog("\nERROR: Unable to write oritatami " + aStr + " to file: '" + fileName + "'\n")
            return False

    def GetAssemblyList(self, bOnlySeed=False):
        assemblyList = []
        if bOnlySeed == True:
            assemblyList = self.assembly.seed
        else:
            for i in range(self.assembly.minX, self.assembly.maxX+1):
                for j in range(self.assembly.minY, self.assembly.maxY+1):
                    locList = self.GetLocOutputList((i,j))
                    if not locList is None:
                        assemblyList.append(locList)
        return assemblyList

    def GetLocOutputList(self, loc):
        onode = self.assembly.GetLoc(loc)
        if onode is not None:
            #self.assembly.ONodeTypes[onode.type]
            locList = []
            locList.append(onode.type)
            locList.append(loc)
            if onode.parent is None:
                locList.append(None)
            else:
                locList.append(onode.parent.coords)
            locList.append(onode._bonds)
            return locList
        else:
            return None
        
    def DoSaveAll(self):
        if self.DoSaveSystemDef(bSaveAs=False) == True:
            if self.DoSaveAssembly(bOnlySeed=True, bSaveAs=False) == True:
                return True
        return False

    def DoSaveAsSVG(self):
        saveFileDialog = wx.FileDialog(self, "Save Assembly As", "", "", 
                                       "SVG files (*.svg)|*.svg",
                                       wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        if not saveFileDialog.ShowModal() == wx.ID_OK:
            return False
        fileName = saveFileDialog.GetPath()
        saveFileDialog.Destroy()
        
        dwg = svgwrite.Drawing(filename=fileName, debug=True)

        #dwg.add(dwg.line((0*mm, 0*mm), (100*mm, 0*mm), stroke=svgwrite.rgb(10, 10, 16, '%')))
        #dwg.add(dwg.text('Test', insert=(0, 0.2), fill='red'))
        #gridGroup.add(dwg.line((startPt[0], startPt[1]), (endPt[0], endPt[1]), stroke=svgwrite.rgb(10, 10, 16, '%')))
        
        self.DrawGrid(None, dwg=dwg)

        dwg.save()
        
    def DoNew(self):
        self.CreateNewAssembly()

        self.SetRandomSeed(self.randSeed)
        SimUtils.globalApp.seedDefinitionFile = None
        SimUtils.globalApp.assemblyDefinitionFile = None
        SimUtils.globalApp.systemDefinitionFile = None
        SimUtils.globalApp.systemDefinitionFileFullPath = None
        
        self.drawingModeBeadType = None
        #self.InitializeDrawing(self.pdc)
        self.app.AddBeadTypeWindow()
        self.app.DrawBeadCycle()
        self.Refresh()
        self.ScrollToCoords((0,0))

        return False

    def Reset(self):
        dlg = wx.TextEntryDialog(self, "Are you sure you want to reset your system to the\nseed assembly and abandon the current assembly?\nIf so, you can set your random seed below\n(use -1 to choose a new random value):", value=str(self.randSeed))
        if not dlg.ShowModal() == wx.ID_OK:
            return
        try:
            result = int(dlg.GetValue())
            self.SetRandomSeed(result)
        except ValueError:
            SimUtils.globalApp.ShowErrorDialog("Random seed value must be an integer.")
            dlg.Destroy()
            return
        dlg.Destroy()
        self.assembly.ResetToSeed()
        self.SetRandomSeed(self.randSeed)
        self.drawnAssembly = []
        self.DrawQueue = []
        #self.InitializeDrawing(self.pdc)
        self.app.DrawBeadCycle()
        self.Refresh()
        self.ScrollToCoords((0,0))
        self.CanvasSizeUpdate(centerCoords=(0,0))
        #self.Refresh()

    def Reload(self):
        dlg = wx.TextEntryDialog(self, "Are you sure you want to reload your system\nand abandon the current assembly?\nIf so, you can set your random seed below\n(use -1 to choose a new random value):", value=str(self.randSeed))
        if not dlg.ShowModal() == wx.ID_OK:
            return
        try:
            result = int(dlg.GetValue())
            self.SetRandomSeed(result)
        except ValueError:
            SimUtils.globalApp.ShowErrorDialog("Random seed value must be an integer.")
            dlg.Destroy()
            return
        dlg.Destroy()

        self.CreateNewAssembly()
        if not SimUtils.globalApp.seedDefinitionFile is None:
            self.assembly.LoadOritatamiAssembly(SimUtils.globalApp.seedDefinitionFile, True)
        elif not SimUtils.globalApp.systemDefinitionFile is None:
            self.assembly.LoadOritatamiSystem(SimUtils.globalApp.systemDefinitionFile, SimUtils.globalApp.systemDefinitionFileFullPath)  
        #self.InitializeDrawing(self.pdc)
        self.app.DrawBeadCycle()
        self.Refresh()
        
    def SetBreakpoints(self):
        outStr = ""
        if self.breakOnNondeterminism == True:
            outStr = "NONDET"
            if len(self.breakpoints) > 0:
                outStr = outStr + ","
        for i in range(len(self.breakpoints)):
            outStr = outStr + str(self.breakpoints[i])
            if i <  (len(self.breakpoints)-1):
                outStr = outStr + ","
        dlg = wx.TextEntryDialog(self, "Please enter a comma separated list of numbers to break\nwhen those numbers of beads are contained in the assembly.\nUse 'NONDET' to break when nondeterminism is detected.", caption="Set Breakpoints", value=str(outStr))
        if not dlg.ShowModal() == wx.ID_OK:
            return
        result = dlg.GetValue()
        dlg.Destroy()
        rawBPs = [x.strip() for x in result.split(',')]
        self.breakOnNondeterminism = False
        self.breakpoints = []
        self.currentBreakpoint = -1
        for bp in rawBPs:
            if len(bp) > 0:
                if bp == "NONDET":
                    self.breakOnNondeterminism = True
                else:
                    try:
                        newBP = int(bp)
                        self.breakpoints.append(newBP)
                    except ValueError:
                        dlg = wx.MessageDialog(self, "Invalid breakpoints entered", caption="Error", style=wx.ICON_ERROR|wx.OK)
                        dlg.ShowModal()
                        dlg.Destroy()
                        self.breakpoints = []
                        self.breakOnNondeterminism = False 
                        return
        if len(self.breakpoints) > 0:
            self.breakpoints.sort()
            self.currentBreakpoint = 0
            
    def DoGo(self):
        if self.keepGoing == True:
            self.app.OutputMessage("ERROR: Cannot create another new thread, there is already one running!")
        self.keepGoing = True
        try:
            self.frame.SetStatusText("", 2)
            self.goThread = thread.start_new_thread(self.GoForward, ())
        except:
            self.app.OutputMessage("ERROR: Failed to create thread!")

    def GoForward(self):
        self.showMouseOver = False
        self.SendStatusBarUpdate(0, "")
        while self.keepGoing == True:
            if self.DoForwardStep() == False:
                self.keepGoing = False
        SimUtils.PrintDebug("Stopping the thread!", SimUtils.MSG_MSG)
        self.showMouseOver = True
        self.SendStatusBarUpdate(0, "Number of beads: " + str(self.assembly.GetNumBeads()))

    def AddToDrawQueue(self, addedList):
        self._lock.acquire()
        for coords in addedList:
            self.DrawQueue.append(coords)
            self.drawnAssembly.append((coords[0], coords[1], self.assembly.GetLoc(coords), self.assembly))
        self._lock.release()
        self.DrawAssemblyUpdate("new")

    def DoForwardStep(self):
        if self.assembly.bAllDone == True:
            return False
        oldgridminx = self.assembly.minX
        coords = self.GetCenterCoords()

        self._lock.acquire()
        (addedStat, addedList) = self.assembly.StepForward()
        self._lock.release()
        if (addedStat == ORI_STAT_NO_SPACE) or (addedStat == ORI_STAT_MAX_FLOPPY) or (addedStat == ORI_STAT_ERROR):
            self.assembly.bAllDone = True
            return False
        else:
            self.AddToDrawQueue(addedList)
        if not self.assembly.minX == oldgridminx:
            if self.drawUpdates == SIMFW_FULLDISPLAY:
                self.UpdateSizes()
                #self.ScrollToCoords(coords)
                self.CanvasSizeUpdate(coords)
            
        #self._lock.acquire()
        self.app.DrawBeadCycle()
        #self._lock.release()
        
        bForce = False
        if (self.assembly.GetNumBeads() % 1000) == 0:
            bForce = True
        self.SendStatusBarUpdate(0, "Number of beads: " + str(self.assembly.GetNumBeads()), bForce)

        if addedStat == ORI_STAT_NONDET and SimUtils.globalApp.canvas.breakOnNondeterminism == True:
            self.app.OutputMessage("Breakpoint encountered due to nondeterminism")
            return False
        elif (self.currentBreakpoint > -1) and (self.assembly.GetNumBeads() >= self.breakpoints[self.currentBreakpoint]):
            self.app.OutputMessage("Breakpoint encountered: " + str(self.breakpoints[self.currentBreakpoint]) + " beads met or exceeded for the first time")
            self.currentBreakpoint = self.currentBreakpoint + 1
            if self.currentBreakpoint >= len(self.breakpoints):
                self.currentBreakpoint = -1
            while (self.currentBreakpoint > -1) and (self.breakpoints[self.currentBreakpoint] < self.assembly.GetNumBeads()):
                self.currentBreakpoint = self.currentBreakpoint + 1
                if self.currentBreakpoint >= len(self.breakpoints):
                    self.currentBreakpoint = -1
            return False
        return True
        
    def DoQuit(self):
        self.keepGoing = False
        self.showMouseOver = True
        self.SendStatusBarUpdate(0, "Number of beads: " + str(self.assembly.GetNumBeads()))
    
    def OnPaint(self, event):
        self._lock.acquire()

        dc = wx.PaintDC(self)

        bg = wx.Brush("WHITE")#(self.GetBackgroundColour())
        dc.SetBackground(bg)
        dc.Clear()

        # use PrepareDC to set position correctly
        self.PrepareDC(dc)
        # we need to clear the dc BEFORE calling PrepareDC
        
        dc.SetUserScale(self.drawScale, self.drawScale)
        
        self.DrawGrid(dc)
        
        for (i,j,loc,ass) in self.drawnAssembly:
            self.DrawAtCoords(dc, i, j, loc, ass)
                
        self._lock.release()
        self.SendStatusBarUpdate(0, "Number of beads: " + str(self.assembly.GetNumBeads()))
        return

    def ZoomThenRecenter(self, zoomIn=True):
        coords = self.GetCenterCoords()
        self.DoScale(zoomIn, scrollToCoords=coords)

    def OnChar(self, evt):
        keycode = evt.GetKeyCode()
        if keycode == ord("a") or keycode == ord("A"):
            self.AddTileCopiesToDrawing()
            self.Refresh()
        elif keycode == ord("f") or keycode == ord("F"):
            self.DoForwardStep()
        elif keycode == ord("p") or keycode == ord("P"):
            self.PrintDebugInfo()
        elif keycode == ord("r") or keycode == ord("R"):
            self.Reset()
        elif keycode == ord("s") or keycode == ord("S"):
            self.DoSaveAll()
        elif keycode == ord("h") or keycode == ord("H"):
            self.ShowHelp()
        elif keycode == ord("l") or keycode == ord("L"):
            self.DoLoad(bAssembly=True, bSystem=True)
        elif keycode == ord("=") or keycode == ord("+"):
            self.DoScale(True, bScrollToCenter=True)
        elif keycode == ord("-") or keycode == ord("_"):
            self.DoScale(False, bScrollToCenter=True)
        elif keycode == ord("c") or keycode == ord("C"):
            self.ScrollToCoords((0,0))
        elif keycode == ord("t") or keycode == ord("T"):
            self.ToggleGrid()
        elif keycode == ord("g") or keycode == ord("G"):
            self.DoGo()
        elif keycode == ord("q") or keycode == ord("Q"):
            self.DoQuit()
        elif keycode == ord("m") or keycode == ord("M"):
            self.MessageToggle(not self.app.showMessages)
            self.app.menuView.Check(self.app.itemTM.GetId(), self.app.showMessages)
        elif keycode == ord("d") or keycode == ord("D"):
            self.ToggleDisplay((self.drawUpdates + 1) % 2)
            self.app.menuView.Check(self.app.itemTD.GetId(), self.drawUpdates == SIMFW_FULLDISPLAY)
        elif keycode == ord("b") or keycode == ord("B"):
            self.SetBreakpoints()
        elif keycode == ord("z") or keycode == ord("Z"):
            self.ToggleDebugMsgs()
        elif keycode == wx.WXK_F2:
            self.Refresh()
        #elif keycode == wx.WXK_DELETE or keycode == wx.WXK_BACK:
            #self.DoDelete()

    def ShowHelp(self):
        self.app.OnHelp(None)

    def PrintDebugInfo(self):        
        #self.app.OutputMessage("some stuff")
        return

    def OnMouse(self, event):

        #=======================================================================
        # if event.IsPageScroll():
        #     print("is page scroll")
        # if event.MiddleIsDown():
        #     print("middle is down")
        #     if event.Dragging():
        #         print("Dragging")
        #=======================================================================
            
        self.SetFocus()

        x,y = self.ConvertEventCoords(event)
        
        if self.showMouseOver == True:
            x2,y2 = self.CoordsFromPixelPos(x, y)
    
            if x2 == None:
                self.SendStatusBarUpdate(1, "")
                self.SendStatusBarUpdate(2, "")
            else:
                self.SendStatusBarUpdate(1, "Coords: (" + str(int(x2)) + "," + str(int(y2)) + ")")
                onode = self.assembly.GetLoc((x2,y2))
                if not onode is None:
                    outStr = "Bead type: " + str(onode.type) + " (Binds to " + str(self.assembly.ONodeTypes[onode.type].bondToTypes) + ")"
                    self.SendStatusBarUpdate(2, outStr)
                else:
                    self.SendStatusBarUpdate(2, "")
        
        if event.LeftDown():
            if self.mouseMode == ORIT_MODE_PAN_ZOOM:
                self.SetCursor(wx.Cursor(self.grabbingCursorImg))
                self.dragStart = self.ScreenToClient(wx.GetMousePosition())
                #self.dragStartScrollPos = (self.GetScrollPos(wx.HORIZONTAL), self.GetScrollPos(wx.VERTICAL))

            elif self.mouseMode == ORIT_MODE_DRAW_BEAD:
                if self.drawingModeBeadType is None:
                    SimUtils.globalApp.ShowErrorDialog("A bead type must be selected in the bead type editor window")
                else:
                    x2,y2 = self.CoordsFromPixelPos(x, y)
                    if (x2 is not None) and (y2 is not None):
                        if self.assembly.GetLoc((x2, y2)) is None:
                            parentLoc = None
                            if (self.assembly.lastBeadLoc is not None) and self.assembly.AreNbrLocs((x2, y2), self.assembly.lastBeadLoc) == True:
                                parentLoc = self.assembly.lastBeadLoc
                            self.ManuallyAddNode(self.drawingModeBeadType.name, (x2, y2), parentLoc)
            elif (self.mouseMode == ORIT_MODE_DRAW_LINK) or (self.mouseMode == ORIT_MODE_DRAW_BOND):
                if (self.linkStart is None) or (self.negSelected is not None):
                    xs,ys = self.CoordsFromPixelPos(x, y)
                    if (xs is None) or (ys is None):
                        self.linkStart = None
                    else:
                        self.linkStart = (xs,ys)
                        node = self.assembly.GetLoc(self.linkStart)
                        if self.negSelected is not None:
                            saved = self.negSelected.coords
                            self.negSelected = None
                            self.ManuallyDrawNodes([saved])
                        self.posSelected = node
                        self.ManuallyDrawNodes([self.linkStart])
                else:
                    xe,ye = self.CoordsFromPixelPos(x, y)
                    if self.linkStart[0] == xe and self.linkStart[1] == ye:
                        node = self.assembly.GetLoc(self.linkStart)
                        self.posSelected = None
                        self.ManuallyDrawNodes([self.linkStart])
                        self.linkStart = None
                    else:
                        if (xe is None) or (ye is None):
                            self.linkEnd = None
                        else:
                            self.linkEnd = (xe,ye)
                        if (self.linkStart is not None) and (self.linkEnd is not None):
                            parentNode = self.assembly.GetLoc(self.linkStart)
                            childNode = self.assembly.GetLoc(self.linkEnd)
                            if (parentNode is not None) and (childNode is not None):
                                if self.mouseMode == ORIT_MODE_DRAW_LINK:
                                    if (self.assembly.AreNbrLocs(self.linkStart, self.linkEnd) == True):
                                        if parentNode.child is not None:
                                            SimUtils.globalApp.ShowErrorDialog("The node at " + str(self.linkStart) + " already has a child")
                                        elif childNode.parent is not None:
                                            SimUtils.globalApp.ShowErrorDialog("The node at " + str(self.linkEnd) + " already has a parent")
                                        else:
                                            bAlreadyBonded = False
                                            for n in range(parentNode.NumBonds()):
                                                b = parentNode.GetNthBond(n)
                                                if b == self.linkEnd:
                                                    bAlreadyBonded = True
                                            if bAlreadyBonded == True:
                                                SimUtils.globalApp.ShowErrorDialog("Nodes must not already be bonded to be linked as parent and child")
                                            else:
                                                parentNode.child = childNode
                                                childNode.parent = parentNode
                                                self.posSelected = None
                                                self.ManuallyDrawNodes([self.linkStart, self.linkEnd])
                                                self.linkStart = None
                                                self.linkEnd = None
                                    else:
                                        SimUtils.globalApp.ShowErrorDialog("Nodes must be neighbors to be linked")
                                elif self.mouseMode == ORIT_MODE_DRAW_BOND:
                                    if (self.assembly.AreNbrLocs(self.linkStart, self.linkEnd) == True):
                                        bAlreadyBonded = False
                                        for n in range(parentNode.NumBonds()):
                                            b = parentNode.GetNthBond(n)
                                            if b == self.linkEnd:
                                                bAlreadyBonded = True
                                        if bAlreadyBonded == True:
                                            SimUtils.globalApp.ShowErrorDialog("Nodes are already bonded")
                                        elif (parentNode.child == childNode) or (childNode.child == parentNode):
                                            SimUtils.globalApp.ShowErrorDialog("Nodes cannot be bonded if they are linked as parent and child")
                                        else:
                                            bDoIt = True
                                            numP = parentNode.NumBonds()
                                            if (self.confirmOverArity == True) and (numP >= self.assembly.arity):
                                                msg = "The bead at location " + str(parentNode.coords) + " currently has " + str(numP)
                                                if numP == 1:
                                                    msg = msg + " bond"
                                                else:
                                                    msg = msg + " bonds"
                                                msg = msg + ", and the arity parameter is set to " + str(self.assembly.arity) + ".\nDo you want to create the new bond anyway?" 
                                                dlg = SimUtils.DontShowAgainDialog(self, wx.ID_ANY, "Confirm exceeding arity", msg)
                                                dlg.CenterOnScreen()
                                                retId = dlg.ShowModal()
                                                dlg.Destroy()
                                                if retId == SimUtils.SIMU_ID_OK_NOCHECK:
                                                    self.confirmOverArity = True
                                                elif retId == SimUtils.SIMU_ID_OK_CHECK:
                                                    self.confirmOverArity = False
                                                else:
                                                    bDoIt = False
                                            numC = childNode.NumBonds()
                                            if (bDoIt == True) and (self.confirmOverArity == True) and (numC >= self.assembly.arity):
                                                msg = "The bead at location " + str(childNode.coords) + " currently has " + str(numC)
                                                if numP == 1:
                                                    msg = msg + " bond"
                                                else:
                                                    msg = msg + " bonds"
                                                msg = msg + ", and the arity parameter is set to " + str(self.assembly.arity) + ".\nDo you want to create the new bond anyway?" 
                                                dlg = SimUtils.DontShowAgainDialog(self, wx.ID_ANY, "Confirm exceeding arity", msg)
                                                dlg.CenterOnScreen()
                                                retId = dlg.ShowModal()
                                                dlg.Destroy()
                                                if retId == SimUtils.SIMU_ID_OK_NOCHECK:
                                                    self.confirmOverArity = True
                                                elif retId == SimUtils.SIMU_ID_OK_CHECK:
                                                    self.confirmOverArity = False
                                                else:
                                                    bDoIt = False
                                            if (bDoIt == True) and (self.confirmInvalidBonds == True) and (self.assembly.ONodeTypes[parentNode.type].BondsToNodeType(childNode.type) == False):
                                                msg = "Bead types '" + str(parentNode.type) + "' and '" + str(childNode.type) + "' don't form bonds."
                                                msg = msg + "\nDo you want to create the new bond anyway?" 
                                                dlg = SimUtils.DontShowAgainDialog(self, wx.ID_ANY, "Confirm bond between non-bonding types", msg)
                                                dlg.CenterOnScreen()
                                                retId = dlg.ShowModal()
                                                dlg.Destroy()
                                                if retId == SimUtils.SIMU_ID_OK_NOCHECK:
                                                    self.confirmInvalidBonds = True
                                                elif retId == SimUtils.SIMU_ID_OK_CHECK:
                                                    self.confirmInvalidBonds = False
                                                else:
                                                    bDoIt = False
                                            if bDoIt == True:
                                                self.assembly.AddBond(self.linkStart, self.linkEnd)
                                                self.posSelected = None
                                                self.ManuallyDrawNodes([self.linkStart, self.linkEnd])
                                                self.linkStart = None
                                                self.linkEnd = None
                                    else:
                                        SimUtils.globalApp.ShowErrorDialog("Nodes must be neighbors to be bonded")

        elif event.RightDown():
            if self.mouseMode == ORIT_MODE_PAN_ZOOM:
                #self.SetCursor(wx.Cursor(self.grabbingCursorImg))
                doScale = True
                if (event.AltDown() == True) or (event.ControlDown() == True) or (event.ShiftDown() == True):
                    doScale = False
                x2,y2 = self.CoordsFromPixelPos(x, y, True)
                self.DoScale(doScale, scrollToCoords=(x2,y2))
                sz = self.GetSize()
                self.WarpPointer(sz[0]/2, sz[1]/2)
                #self.SetCursor(wx.StockCursor(wx.CURSOR_ARROW))

            elif self.mouseMode == ORIT_MODE_DRAW_BEAD:
                x2,y2 = self.CoordsFromPixelPos(x, y)
                if (x2 is not None) and (y2 is not None):
                    if self.assembly.GetLoc((x2, y2)) is not None:
                        self.ManuallyRemoveNode((x2, y2))
                        
            elif (self.mouseMode == ORIT_MODE_DRAW_LINK) or (self.mouseMode == ORIT_MODE_DRAW_BOND):
                if (self.linkStart is None) or (self.posSelected is not None):
                    xs,ys = self.CoordsFromPixelPos(x, y)
                    if (xs is None) or (ys is None):
                        self.linkStart = None
                    else:
                        self.linkStart = (xs,ys)
                        node = self.assembly.GetLoc(self.linkStart)
                        if self.posSelected is not None:
                            saved = self.posSelected.coords
                            self.posSelected = None
                            self.ManuallyDrawNodes([saved])
                        self.negSelected = node
                        self.ManuallyDrawNodes([self.linkStart])
                else:
                    xe,ye = self.CoordsFromPixelPos(x, y)
                    if self.linkStart[0] == xe and self.linkStart[1] == ye:
                        node = self.assembly.GetLoc(self.linkStart)
                        self.negSelected = None
                        self.ManuallyDrawNodes([self.linkStart])
                        self.linkStart = None
                    else:
                        if (xe is None) or (ye is None):
                            self.linkEnd = None
                        else:
                            self.linkEnd = (xe,ye)
                        if (self.linkStart is not None) and (self.linkEnd is not None):
                            parentNode = self.assembly.GetLoc(self.linkStart)
                            childNode = self.assembly.GetLoc(self.linkEnd)
                            if (parentNode is not None) and (childNode is not None):
                                bDidIt = False
                                if self.mouseMode == ORIT_MODE_DRAW_LINK:
                                    if (parentNode.child == childNode):
                                        parentNode.child = None
                                        childNode.parent = None
                                        bDidIt = True
                                    elif (childNode.child == parentNode):
                                        childNode.child = None
                                        parentNode.parent = None
                                        bDidIt = True
                                    if bDidIt == True:
                                        self.negSelected = None
                                        self.linkStart = None
                                        self.linkEnd = None
                                        self.Refresh()
                                    else:
                                        SimUtils.globalApp.ShowErrorDialog("Cannot remove link between beads which aren't linked.")
                                elif self.mouseMode == ORIT_MODE_DRAW_BOND:
                                    bAlreadyBonded = False
                                    for n in range(parentNode.NumBonds()):
                                        b = parentNode.GetNthBond(n)
                                        if b == self.linkEnd:
                                            bAlreadyBonded = True
                                    if bAlreadyBonded == True:
                                        self.assembly.RemoveBond(self.linkStart, self.linkEnd)
                                        bDidIt = True
                                    if bDidIt == True:
                                        self.negSelected = None
                                        self.linkStart = None
                                        self.linkEnd = None
                                        self.Refresh()
                                    else:
                                        SimUtils.globalApp.ShowErrorDialog("Cannot remove bond between beads which don't have a bond.")

        if event.LeftUp():
            self.dragStart = None
            self.SetCursor(wx.Cursor(wx.CURSOR_ARROW))

        elif event.Dragging():
            if self.mouseMode == ORIT_MODE_PAN_ZOOM:
                if self.dragStart is not None:
                    new = self.ScreenToClient(wx.GetMousePosition())
                    dx = int((self.dragStart.x-new.x)/1)
                    dy = int((self.dragStart.y-new.y)/1)
                    spx = self.GetScrollPos(wx.HORIZONTAL)
                    spy = self.GetScrollPos(wx.VERTICAL)
                    self.Scroll(spx+dx, spy+dy)
                    self.dragStart = new
            
            elif self.mouseMode == ORIT_MODE_DRAW_BEAD:
                if event.LeftIsDown():
                    if self.drawingModeBeadType is None:
                        SimUtils.globalApp.ShowErrorDialog("A bead type must be selected in the bead type editor window")
                    else:
                        x2,y2 = self.CoordsFromPixelPos(x, y)
                        if (x2 is not None) and (y2 is not None):
                            if self.assembly.GetLoc((x2, y2)) is None:
                                parentLoc = None
                                if (self.assembly.lastBeadLoc is not None) and self.assembly.AreNbrLocs((x2, y2), self.assembly.lastBeadLoc) == True:
                                    parentLoc = self.assembly.lastBeadLoc
                                self.ManuallyAddNode(self.drawingModeBeadType.name, (x2, y2), parentLoc)
                elif event.RightIsDown():
                    x2,y2 = self.CoordsFromPixelPos(x, y)
                    if (x2 is not None) and (y2 is not None):
                        if self.assembly.GetLoc((x2, y2)) is not None:
                            self.ManuallyRemoveNode((x2, y2))

    def ManuallyAddNode(self, typeName, coords, parentCoords):
        self.assembly.AddNode(self.drawingModeBeadType.name, coords, parentCoords)
        self.ManuallyDrawNodes([coords])
        self.SendStatusBarUpdate(0, "Number of beads: " + str(self.assembly.GetNumBeads()))
        
    def ManuallyRemoveNode(self, coords):
        self.assembly.RemoveNode(coords)
        self.Refresh()
        self.SendStatusBarUpdate(0, "Number of beads: " + str(self.assembly.GetNumBeads()))
        
    def ManuallyDrawNodes(self, coordList):
        self._lock.acquire()
        for coords in coordList:
            self.DrawQueue.append(coords)
        self._lock.release()
        self.DrawAssemblyUpdate("new")

    def ScrollToCoords(self, coords):
        if (not coords == None) and (not coords[0] == None) and (not coords[1] == None):
            (x,y) = self.GetPixelsForCoords(coords[0], coords[1])
            
            # Some slight adjustments due to positioning of triangular grid
            x = x - self.sideLen
            y = y - self.sideLen

            x = x * self.drawScale
            y = y * self.drawScale
            (xs, ys) = self.GetScrollPixelsPerUnit()
            (xsz,ysz) = self.GetSize()
            self.Scroll((x-xsz/2+self.sideLen*self.drawScale)/xs, (y-ysz/2+self.sideLen*self.drawScale)/ys)

    def ColoredPen(self, color, width = 2):
        if not self.pen_cache.has_key( (color, width) ):
            self.pen_cache[(color, width)] = wx.Pen(color, width)
        return self.pen_cache[(color, width)]

    def ColoredBrush(self, colorName):
        if not self.brush_cache.has_key(colorName):
            if type(colorName) is str:
                if "(" in colorName:
                    colorTuple = eval(colorName)
                    if len(colorTuple) > 3:
                        color = wx.Colour(colorTuple[0], colorTuple[1], colorTuple[2], colorTuple[3])
                    else:
                        color = wx.Colour(colorTuple[0], colorTuple[1], colorTuple[2])
                else:
                    color = wx.TheColourDatabase.Find(colorName)
                self.brush_cache[colorName] = wx.Brush(color)
        return self.brush_cache[colorName]


#===============================================================================
#     def OnPaint(self, event):
#         # Create a buffered paint DC.  It will create the real
#         # wx.PaintDC and then blit the bitmap to it when dc is
#         # deleted.  
#         dc = wx.BufferedPaintDC(self)
#         # use PrepateDC to set position correctly
#         self.PrepareDC(dc)
#         # we need to clear the dc BEFORE calling PrepareDC
#         bg = wx.Brush(self.GetBackgroundColour())
#         dc.SetBackground(bg)
#         dc.Clear()
# 
#         dc.SetUserScale(self.drawScale, self.drawScale)
# 
#         # create a clipping rect from our position and size
#         # and the Update Region
#         xv, yv = self.GetViewStart()
#         dx, dy = self.GetScrollPixelsPerUnit()
#         x, y   = (xv * dx, yv * dy)
#         rgn = self.GetUpdateRegion()
#         rgn.Offset(x,y)
#         r = rgn.GetBox()
#         # draw to the dc using the calculated clipping rect
#         self.pdc.DrawToDCClipped(dc,r)
#===============================================================================

        
    def DoDrawAssemblyUpdate(self, msg):
        if self.drawUpdates == SIMFW_NODISPLAY:
            return

        if msg == "all":
            self.Refresh()
            return
        elif msg == "allAndCheck":
        #    self.DoInitializeDrawing(dc)
            self.DoRedrawAssembly()

    
        dc = wx.ClientDC(self)
        
        self.PrepareDC(dc)
        # we need to clear the dc BEFORE calling PrepareDC
        dc.SetUserScale(self.drawScale, self.drawScale)

        #print("in drawupdate, queue len: " + str(len(self.DrawQueue)))
        if (msg == "new"):
            
            self._lock.acquire()
            copiedList = list(self.DrawQueue)
            if self.lastCurrent is not None:    # Redraw the bead that was current at the last step
                copiedList.append(self.lastCurrent)
            self.DrawQueue = []
            
            for e in copiedList:
                onode = self.assembly.GetLoc((e[0], e[1]))
                if onode is not None:
                    self.DrawAtCoords(dc, e[0], e[1], onode, self.assembly)
            self._lock.release()
    
            if len(self.RemoveQueue) > 0:
                self.RemoveQueue = []
                self.DoRedrawAssembly()

        
    def UpdateSizes(self):
        # Get new assembly dimensions
        self.gridWidth = self.assembly.GetWidth()
        self.gridHeight = self.assembly.GetHeight()
        
        # Compute new dimensions for canvas
        self.width  = self.sideLen * self.gridWidth  + self.GetHorizOffset(self.gridWidth)
        self.height = self.GetVertOffset(self.gridHeight)
        self.SetVirtualSize((self.width*self.drawScale, self.height*self.drawScale))
        
    def DrawGrid(self, dc, dwg=None):
        self.UpdateSizes()

        if dwg is None:
            #  Remove old data
            dc.Clear()
            dc.SetPen(self.gridPen)
        else:
            grid = dwg.g(id="grid")
            gridGroup = dwg.add(grid)

        if (self.showGrid == SHOW_GRID_AND_COORDS) or (self.showGrid == SHOW_GRID_ONLY):
            # Draw horizontal lines            
            for i in range(self.assembly.minX, self.assembly.maxX+1):
                startPt = self.GetPixelsForCoords(self.assembly.minX, i)
                endPt = self.GetPixelsForCoords(self.assembly.maxX, i)
                if dwg is None:
                    dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])
                else:
                    gridGroup.add(dwg.line((startPt[0], startPt[1]), (endPt[0], endPt[1]), stroke="black"))

            # Draw positive sloped vertical lines
            #MJP: assuming a square grid (i.e. same number vertical and horizontal cells)
            for i in range(self.assembly.minX, self.assembly.maxX+1):
                startPt = self.GetPixelsForCoords(i, self.assembly.minY)
                endPt = self.GetPixelsForCoords(i, self.assembly.maxY)
                if dwg is None:
                    dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])
                else:
                    gridGroup.add(dwg.line((startPt[0], startPt[1]), (endPt[0], endPt[1]), stroke="black"))

            # Draw negative sloped vertical lines
            #MJP: assuming a square grid (i.e. same number vertical and horizontal cells)
            for i in range(self.assembly.minX, self.assembly.maxX+1):
                startPt = self.GetPixelsForCoords(i, self.assembly.minY)
                endPt = self.GetPixelsForCoords(self.assembly.minX,i)
                if dwg is None:
                    dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])
                else:
                    gridGroup.add(dwg.line((startPt[0], startPt[1]), (endPt[0], endPt[1]), stroke="black"))
                startPt = self.GetPixelsForCoords(i, self.assembly.maxY)
                endPt = self.GetPixelsForCoords(self.assembly.maxX, i)
                if dwg is None:
                    dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])
                else:
                    gridGroup.add(dwg.line((startPt[0], startPt[1]), (endPt[0], endPt[1]), stroke="black"))

        if self.showGrid == SHOW_GRID_AND_COORDS:
            for i in range(self.assembly.minX, self.assembly.maxX+1):
                for j in range(self.assembly.minY, self.assembly.maxY+1):
                    (xp,yp) = self.GetPixelsForCoords(i,j)
                    if dwg is None:
                        dc.DrawText(str(i) + "," + str(j), xp+8, yp-2)
                    else:
                        gridGroup.add(dwg.text(str(i) + "," + str(j), font_size=8.5, insert=(xp+7.5, yp+9.5)))

        for i in range(self.assembly.minX, self.assembly.maxX+1):
            for j in range(self.assembly.minY, self.assembly.maxY+1):
                onode = self.assembly.GetLoc((i,j))
                if onode is not None:
                    self.DrawAtCoords(dc, i, j, onode, self.assembly, dwg)

    def GetHorizOffset(self, y):
        return y * self.sideLen * math.cos(math.radians(60))
    
    def GetVertOffset(self, y):
        return y * self.sideLen * math.cos(math.radians(30))

    def GetPixelsForCoords(self, xCoord, yCoord):
        xCoordAdj = xCoord - self.assembly.minX
        yCoordAdj = yCoord - self.assembly.minY
        xPixel = (xCoordAdj * self.sideLen) + self.GetHorizOffset(yCoordAdj)
        yPixel = self.GetVertOffset(yCoordAdj)        
        yPixel = self.height - (yPixel + self.sideLen/2)
        return (xPixel, yPixel)

    def CoordsFromPixelPos(self, xPixel, yPixel, bClosest=False):
        sz = self.GetVirtualSize()
        yPixel = sz[1] - (yPixel + (self.sideLen/2)*self.drawScale)
        xPixel = xPixel / self.drawScale
        yPixel = yPixel / self.drawScale
        yCoord = (yPixel / (self.sideLen * math.cos(math.radians(30)))) + self.assembly.minX
        xCoord = ((xPixel - self.GetHorizOffset(yCoord - self.assembly.minY)) / self.sideLen) + self.assembly.minX
        (x2,x1) = math.modf(xCoord)
        (y2,y1) = math.modf(yCoord)
        if ((abs(x2) < 0.2 or abs(x2) > 0.8) and (abs(y2) < 0.2 or abs(y2) > 0.8)) or (bClosest == True):
            retX = int(round(xCoord))
            retY = int(round(yCoord))
            if (retX >= self.assembly.minX and retX <= self.assembly.maxX and retY >= self.assembly.minY and retY <= self.assembly.maxY):
                return (retX, retY)
        return (None,None)

    def DrawAtCoords(self, dc, xCoord, yCoord, onode, grid, dwg=None):
        
        (xPixel,yPixel) = self.GetPixelsForCoords(xCoord, yCoord)

        if not dwg is None:
            bGroup = dwg.add(dwg.g(id="b_" + str(xCoord) + "_" + str(yCoord)))
            
        if not onode.parent is None:
            (xPixelP,yPixelP) = self.GetPixelsForCoords(onode.parent.coords[0], onode.parent.coords[1])
            if dwg is None:
                dc.SetPen(self.linkPen)
            
            # Draw parent-child bond
            if dwg is None:
                dc.DrawLine(xPixel,yPixel,xPixelP,yPixelP)
            else:
                bGroup.add(dwg.line((xPixel,yPixel), (xPixelP,yPixelP), stroke_width=3, stroke="black"))
            
            # Draw directional 'arrow' lines
            frac = 0.1
            midpoint = (int((xPixel+xPixelP)/2), int((yPixel+yPixelP)/2))
            slope = (xPixel-xPixelP, yPixel-yPixelP)
            midpoint = (midpoint[0]+slope[0]*frac, midpoint[1]+slope[1]*frac)
            invSlope = ((self.sideLen-abs(slope[0]))/self.sideLen, (self.sideLen-abs(slope[1]))/self.sideLen)

            linepoint = (int(midpoint[0]-slope[0]*frac), int(midpoint[1]-slope[1]*frac))

            mult = 10
            if (slope[0] == 0) or (slope[1] == 0):  # Horizontal
                pt1 = (int(linepoint[0]+invSlope[0]*7), int(linepoint[1]-invSlope[1]*7))
                pt2 = (int(linepoint[0]-invSlope[0]*7), int(linepoint[1]+invSlope[1]*7))
            elif ((slope[0] < 0) and (slope[1] < 0)) or ((slope[0] > 0) and (slope[1] > 0)):
                pt1 = (int(linepoint[0]+invSlope[0]*mult), int(linepoint[1]-invSlope[1]*mult))
                pt2 = (int(linepoint[0]-invSlope[0]*mult), int(linepoint[1]+invSlope[1]*mult))
            else:
                pt1 = (int(linepoint[0]+invSlope[0]*mult), int(linepoint[1]+invSlope[1]*mult))
                pt2 = (int(linepoint[0]-invSlope[0]*mult), int(linepoint[1]-invSlope[1]*mult))
            
            if dwg is None:
                dc.DrawLine(midpoint[0], midpoint[1], pt1[0], pt1[1])
                dc.DrawLine(midpoint[0], midpoint[1], pt2[0], pt2[1])
            else:
                bGroup.add(dwg.line((midpoint[0], midpoint[1]), (pt1[0], pt1[1]), stroke_width=3, stroke="black"))
                bGroup.add(dwg.line((midpoint[0], midpoint[1]), (pt2[0], pt2[1]), stroke_width=3, stroke="black"))

        numBonds = onode.NumBonds()
        if numBonds > 0:
            if dwg is None:
                dc.SetPen(self.bondPen)
            for n in range(numBonds):
                b = onode.GetNthBond(n)
                (x2, y2) = self.GetPixelsForCoords(onode.coords[0], onode.coords[1])
                (x3, y3) = self.GetPixelsForCoords(b[0], b[1])
                if dwg is None:
                    dc.DrawLine(x2, y2, x3, y3)
                else:
                    bGroup.add(dwg.line((x2, y2), (x3, y3), stroke_width=2.0, stroke="red"))
        if dwg is None:
            dc.SetPen(self.gridPen)

        if onode.type in grid.ONodeTypes:
            nodeColor = grid.ONodeTypes[onode.type].color
        else:
            nodeColor = "black"
        pointBrush = self.ColoredBrush(nodeColor)
        if dwg is None:
            dc.SetBrush(pointBrush)
            dc.DrawCircle((xPixel,yPixel), 10)
        else:
            pbc = pointBrush.GetColour()
            col = svgwrite.rgb(pbc.Red(), pbc.Green(), pbc.Blue(), mode="RGB")
            bGroup.add(dwg.circle((xPixel,yPixel), 10, stroke="black", fill=col))

        if onode == self.posSelected:
            pointBrush = self.ColoredBrush("Green")
            if dwg is None:
                dc.SetBrush(pointBrush)
                dc.DrawCircle((xPixel,yPixel), 5)
            else:
                bGroup.add(dwg.circle((xPixel,yPixel), 5, stroke="green"))
        elif onode == self.negSelected:
            pointBrush = self.ColoredBrush("Red")
            if dwg is None:
                dc.SetBrush(pointBrush)
                dc.DrawCircle((xPixel,yPixel), 5)
            else:
                bGroup.add(dwg.circle((xPixel,yPixel), 5, stroke="red"))
        elif onode.coords == self.assembly.lastBeadLoc:
            pointBrush = self.ColoredBrush("White")
            if dwg is None:
                dc.SetBrush(pointBrush)
                dc.DrawCircle((xPixel,yPixel), 5)
            else:
                bGroup.add(dwg.circle((xPixel,yPixel), 5, stroke="white", fill="white"))
            self.lastCurrent = onode.coords
        if not dwg is None:
            dwg.add(bGroup)
        
    def ToggleGrid(self):
            self.showGrid = (self.showGrid + 1) % 3
            self.InitializeDrawing()

    def ToggleDisplay(self, showMode):
        self.drawUpdates = showMode
        if self.drawUpdates == SIMFW_FULLDISPLAY:
            self.app.OutputMessage("Setting display mode to display all assembly updates.")
            self.Refresh()
            if not self.savedCenteredCoords is None:
                self.ScrollToCoords(self.savedCenteredCoords)
        else:
            self.savedCenteredCoords = self.GetCenterCoords()
            self.app.OutputMessage("Setting display mode to hide all assembly updates.")

    def ToggleDebugMsgs(self):
        self.app.bShowDebugMsgs = not self.app.bShowDebugMsgs
        if self.app.bShowDebugMsgs == True:
            self.app.itemShowDebug.Check(True)
            self.app.OutputMessage("Set to display debugging messages.")
        else:
            self.app.itemShowDebug.Check(False)
            self.app.OutputMessage("Set to hide debugging messages.")

    def MessageToggle(self, bShow):
        if bShow == True:
            self.app.showMessages = bShow
            self.app.OutputMessage("Setting display mode to display all message and bead sequence panel updates.")
            self.app.DrawBeadCycle()
        else:
            self.app.OutputMessage("Setting display mode to hide all message and bead sequence panel updates.")
            self.app.showMessages = bShow
