import SimUtils
import TileAssembly
import Frontier
import sys
import random
import thread
import threading
import wx
import wx.adv
from wx.lib.pubsub import pub
import SimCanvasRubberband


CELLSIZE  = 60 #96
IMG_OFFSET_HORIZ = 17
IMG_OFFSET_VERT  = 17

SIMFW_FULLDISPLAY      = 0
#SIMFW_GROWTHDISPLAY    = 1
SIMFW_NODISPLAY        = 1

SIMFW_MODE_ZOOM   = 0
SIMFW_MODE_SELECT = 1


class Canvas(wx.ScrolledWindow):
    def __init__(self, parent, idIn, frame, size, app):
        wx.ScrolledWindow.__init__(self, parent, idIn, (0, 0), size=size, style=wx.SUNKEN_BORDER)

        self.frame = frame
        self.app = app
        self.lines = []
        self.CreateNewAssembly()
        self.width  = self.assembly.GetWidth() * CELLSIZE
        self.height = self.assembly.GetHeight() * CELLSIZE
        self.sideLen = CELLSIZE
        self.drawScale = 1.0
        self.DrawQueue = []
        self.RemoveQueue = []
        self.drawGrid = False
        self.keepGoing = False
        self.drawUpdates = SIMFW_FULLDISPLAY
        self.showMouseOver = True
        self.displayPatterns = []
        self.drawnAssembly = []
        self.breakpoints = []
        self.currentBreakpoint = -1
        self._lock = threading.Lock()
        self.breakOnNondeterminism = False
        self.savedCenteredCoords = None
        
        #Can't do this!  It causes a full redraw (i.e. paint) after every draw
        #self.SetDoubleBuffered(True)

        self.SetRandomSeed()
        
        self.numVert = self.width / self.sideLen
        self.numHoriz = self.height / self.sideLen

        # create a PseudoDC to record our drawing
        #self.pdc = wx.PseudoDC()

        self.pen_cache = {}
        self.brush_cache = {}
        self.gridPen = self.ColoredPen("BLACK", 1)
        self.SetBackgroundColour("WHITE")

        self.SetVirtualSize((self.width*self.drawScale, self.height*self.drawScale))
        self.SetScrollRate(20,20)

        self.lastRect = wx.Rect(-999999999, -999999999, 0, 0)
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_ERASE_BACKGROUND, lambda x:None)
        self.Bind(wx.EVT_MOUSE_EVENTS, self.OnMouse)
        self.Bind(wx.EVT_CHAR,self.OnChar)
        
        pub.getDefaultPublisher().subscribe(self.UpdateStatusbar, "updateStatusbar")
        pub.getDefaultPublisher().subscribe(self.DoDrawAssemblyUpdate, "drawAssembly")
        pub.getDefaultPublisher().subscribe(self.DoCanvasSizeUpdate, "sizeCanvas")

        self.mouseMode = SIMFW_MODE_ZOOM

        self.InitializeDrawing()
        #self.Refresh()
        self.ScrollToCoords((0,0))
        
        self.rubberBand = SimCanvasRubberband.SimCanvasRubberband(drawingSurface=self)
        self.rubberBand.reset(aspectRatio=0.5)
        self.rubberBand.reset()
        self.lastDeletedList = []

    def CreateNewAssembly(self):
        self.assembly = TileAssembly.Assembly()

    def SendStatusBarUpdate(self, field, msg):
        wx.CallAfter(pub.getDefaultPublisher().sendMessage, "updateStatusbar", msg = (field, msg))

    def UpdateStatusbar(self, msg):
        self.frame.SetStatusText(msg[1], msg[0])

    def SetRandomSeed(self, seed=-1):
        self.randSeed = seed
        if self.randSeed < 0:
            self.randSeed = random.randint(0, sys.maxint)
        SimUtils.globalApp.OutputMessage("Seeding random number generator with " + str(self.randSeed))
        random.seed(self.randSeed)

    def DoLoad(self):
        openFileDialog = wx.FileDialog(self.app.window, "Open", "", "", 
                                       "Tile system files (*.txt)|*.txt",
                                       wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
        openFileDialog.ShowModal()
        fileName = openFileDialog.GetPath()
        openFileDialog.Destroy()
        if (len(fileName) > 0):
            
            dlg = wx.TextEntryDialog(self, "You can set your random seed below\n(use -1 to choose a new random value):", caption="Set random seed", value=str(self.randSeed))
            if dlg.ShowModal() == wx.ID_OK:
                self.randSeed = int(dlg.GetValue())
                dlg.Destroy()
            else:
                dlg.Destroy()
                return

            self.SetRandomSeed(self.randSeed)
            SimUtils.globalApp.OutputMessage("Loading tile assembly system...")
            self.InitializeDrawing()
            self.CreateNewAssembly()
            self.assembly.LoadTileSystem(fileName)
            self.RedrawAssembly(checkAll=True)
            self.ScrollToCoords((0,0))
            SimUtils.globalApp.OutputMessage("Tile assembly system loaded")

    def DoSave(self):
        #saveFileDialog = wx.FileDialog(self.window, "Save As", "", "", 
        #                               "Tile system files (*.txt)|*.txt",
        #                               wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        #saveFileDialog.ShowModal()
        #fileName = saveFileDialog.GetPath()
        #saveFileDialog.Destroy()

        #self.saveSnapshot()

        dlg = wx.MessageDialog(self, "Sorry, this hasn't been implemented yet..")
        dlg.ShowModal()
        dlg.Destroy()
        return
    
        currFile = "output.txt"
        newFile = wx.FileSelector("Save assembly to file...", currFile, currFile, ".txt", "*.txt", wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
        if len(newFile) > 0:
            fOut = open(newFile, "w")
            iCount = 0
            jCount = 0
            for i in range(self.numHoriz):
                for j in range(self.numVert):
                    if self.grid.IsEmpty(i, j) == False:
                        pt = self.grid.GetCenterOffsetOfType(i, j)
                        if pt.x == 0 and pt.y == 0:
                            name = self.grid.GetType(i, j)
                            outStr = name + "," + str(i) + "," + str(j) + "\n"
                            fOut.write(outStr)
            fOut.close()

    def saveSnapshot(self, dcSource):
        
        dlg = wx.MessageDialog(self, "Sorry, this hasn't been implemented yet..")
        dlg.ShowModal()
        dlg.Destroy()
        return

        # based largely on code posted to wxpython-users by Andrea Gavana 2006-11-08
        #size = dcSource.Size
    
        # Create a Bitmap that will later on hold the screenshot image
        # Note that the Bitmap must have a size big enough to hold the screenshot
        # -1 means using the current default colour depth
#         bmp = wx.EmptyBitmap(self.width, self.height)
#     
#         # Create a memory DC that will be used for actually taking the screenshot
#         memDC = wx.MemoryDC()
#     
#         # Tell the memory DC to use our Bitmap
#         # all drawing action on the memory DC will go to the Bitmap now
#         memDC.SelectObject(bmp)
#     
#         # Blit (in this case copy) the actual screen on the memory DC
#         # and thus the Bitmap
#         memDC.Blit( 0, # Copy to this X coordinate
#             0, # Copy to this Y coordinate
#             self.width, # Copy this width
#             self.height, # Copy this height
#             dcSource, # From where do we copy?
#             0, # What's the X offset in the original DC?
#             0  # What's the Y offset in the original DC?
#             )
        
        bmp = wx.EmptyBitmap(self.width, self.height)
        #wdc = wx.WindowDC(self)
        memDC = wx.MemoryDC(bmp)

        bg = wx.Brush("WHITE")
        memDC.SetBackground(bg)
        memDC.Clear()
        #memDC.Blit(0, 0, self.width, self.height, wdc, 0, 0)
        
        for i in range(self.assembly.minX, self.assembly.maxX+1):
            for j in range(self.assembly.minY, self.assembly.maxY+1):
                loc = self.assembly.GetLoc((i,j))
                if loc is not None:
                    if self.NameMatchesPatterns(loc.type.name) == True:
                        self.DrawAtCoords(memDC, i, j, self.assembly.GetLoc((i,j)))
        
        # Select the Bitmap out of the memory DC by selecting a new
        # uninitialized Bitmap
        memDC.SelectObject(wx.NullBitmap)
        imageFileName = 'snapshots-saved.bmp'
        bmp.SaveFile(imageFileName, wx.BITMAP_TYPE_BMP)
        
        ##pdfFileName = 'snapshots-saved.pdf'
        #=======================================================================
        # c = canvas.Canvas(pdfFileName, pagesize=letter)
        # width, height = letter
        # c.drawImage(imageFileName, inch, height - 2 * inch) # Who needs consistency?
        # c.showPage()
        # c.save()
        #=======================================================================
        #=======================================================================
        # doc = SimpleDocTemplate(pdfFileName, pagesize=(self.width,self.height))
        # parts = []
        # parts.append(Image(imageFileName))
        # doc.build(parts)
        #=======================================================================
    
        #img = bmp.ConvertToImage()
        #img.SaveFile('snapshots-saved.png', wx.BITMAP_TYPE_PNG)

    def Reset(self, bFromFiles=True):
        dlg = wx.TextEntryDialog(self, "Are you sure you want to reset your system\nand abandon the current assembly?\nIf so, you can set your random seed below\n(use -1 to choose a new random value):", caption="Set random seed", value=str(self.randSeed))
        if not dlg.ShowModal() == wx.ID_OK:
            return
        result = int(dlg.GetValue())
        dlg.Destroy()
        self.SetRandomSeed(result)

        self.app.OutputMessage("Reseting system...")
        
        if bFromFiles == True:
            self.CreateNewAssembly()
            if not self.systemDefFile is None:
                self.assembly.LoadTileSystem(self.systemDefFile)
        else:
            self.assembly.ClearSpace()
            self.assembly.PlaceSeed()
            
        self.width = self.assembly.width * CELLSIZE
        self.height = self.assembly.height * CELLSIZE
        #self.SetVirtualSize((self.width*self.drawScale, self.height*self.drawScale))
        self.RedrawAssembly(checkAll=True)
        self.ScrollToCoords((0,0))
        self.app.OutputMessage("Tile assembly system reset")
        

    def DoGo(self):
        if self.keepGoing == True:
            self.app.OutputMessage("ERROR: Cannot create another new thread, there is already one running!")
        self.keepGoing = True
        try:
            self.goThread = thread.start_new_thread(self.GoForward, ())
        except:
            self.app.OutputMessage("ERROR: Failed to create thread!")
    
    def DoQuit(self):
        self.keepGoing = False
        self.showMouseOver = True
        self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles))
    
    def GoForward(self):
        self.showMouseOver = False
        self.SendStatusBarUpdate(1, "")
        while self.keepGoing == True:
            if self.DoForwardStep() == False:
                self.keepGoing = False
        SimUtils.PrintDebug("Stopping the thread!", SimUtils.MSG_MSG)
        self.showMouseOver = True
        self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles))

    def DoForwardStep(self):
        newLoc = self.assembly.StepForward()
        if not newLoc is None:
            if not self.width == (self.assembly.width * CELLSIZE) or not self.height == (self.assembly.height * CELLSIZE):
                coords = self.GetCenterCoords()
                self.width = self.assembly.width * CELLSIZE
                self.height = self.assembly.height * CELLSIZE
                self.numVert = self.width / self.sideLen
                self.numHoriz = self.height / self.sideLen
                self.RedrawAssembly()
                self.CanvasSizeUpdate(coords)
            else:
                self.AddToDrawQueue(newLoc)
                    #self.DrawUpdate()
                    #self.Refresh()
            #print("Number of tiles: " + str(self.assembly.numTiles))
            if self.drawUpdates == SIMFW_NODISPLAY:
                if (self.assembly.numTiles % 5000) == 0:
                    self.app.OutputMessage("Number of tiles in assembly: " + str(self.assembly.numTiles))
            elif self.keepGoing == False:
                #MJP TODO: handle this better, but it REALLY slows things down!!!
                self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles))
            return True
        return False

    def GetCenterCoords(self):
        usp = self.CalcUnscrolledPosition(0,0)
        sz = self.GetSize()
        center = (usp[0] + sz[0]/2, usp[1] + sz[1]/2)
        coords = self.CoordsFromPixelPos(center[0], center[1], True)
        return coords
    
    def ConvertEventCoords(self, event):
        xView, yView = self.GetViewStart()
        xDelta, yDelta = self.GetScrollPixelsPerUnit()
        return (event.GetX() + (xView * xDelta),
            event.GetY() + (yView * yDelta))

    def OffsetRect(self, r):
        xView, yView = self.GetViewStart()
        xDelta, yDelta = self.GetScrollPixelsPerUnit()
        r.OffsetXY(-(xView*xDelta),-(yView*yDelta))

    def DeleteSelection(self):
        xView, yView = self.GetViewStart()
        xDelta, yDelta = self.GetScrollPixelsPerUnit()

        (x1,y1,x2,y2) = self.rubberBand.currentBox
        (xLeft, yTop) = (x1 + (xView * xDelta), y1 + (yView * yDelta))
        (xRight, yBottom) = (x1 + x2 + (xView * xDelta), y1 + y2 + (yView * yDelta))

        cLeft, cTop = self.CoordsFromPixelPos(xLeft, yTop)
        cRight, cBottom = self.CoordsFromPixelPos(xRight, yBottom)

        self.lastDeletedList = []
        for x in range(cLeft+1, cRight):
            for y in range(cBottom+1, cTop):
                loc = self.assembly.GetLoc((x,y))
                if not loc is None:
                    if self.assembly.RemoveTile((x,y)) == True:
                        self.lastDeletedList.append(loc)
        self.DoRedrawAssembly()
        
    def MessageToggle(self, bShow):
        if bShow == True:
            self.app.showMessages = bShow
            self.app.OutputMessage("Setting display mode to display all messages.")
        else:
            self.app.OutputMessage("Setting display mode to hide all message updates.")
            self.app.showMessages = bShow

    def PrintMessage(self):
        print("This is a placeholder for printing out debugging or other useful info")

    def OnChar(self, evt):
        keycode = evt.GetKeyCode()
        if keycode == ord("f") or keycode == ord("F"):
            self.DoForwardStep()
        elif keycode == ord("p") or keycode == ord("P"):
            self.PrintMessage()
        elif keycode == ord("r") or keycode == ord("R"):
            self.Reset(False)
        elif keycode == ord("s") or keycode == ord("S"):
            self.DoSave()
        elif keycode == ord("h") or keycode == ord("H"):
            self.ShowHelp()
        elif keycode == ord("l") or keycode == ord("L"):
            self.DoLoad()
        elif keycode == ord("=") or keycode == ord("+"):
            self.ZoomThenRecenter(True)
        elif keycode == ord("-") or keycode == ord("_"):
            self.ZoomThenRecenter(False)
        elif keycode == ord("g") or keycode == ord("G"):
            self.DoGo()
        elif keycode == ord("q") or keycode == ord("Q"):
            self.DoQuit()
        elif keycode == ord("c") or keycode == ord("C"):
            self.ScrollToCoords((0,0))
        elif keycode == ord("t") or keycode == ord("T"):
            self.ToggleGrid()
        elif keycode == ord("d") or keycode == ord("D"):
            self.ToggleDisplay()
        elif keycode == wx.WXK_F2:
            self.Refresh()
        elif keycode == wx.WXK_RETURN:
            self.DeleteSelection()
        elif keycode == wx.WXK_F2:
            self.Refresh()
        #elif keycode == wx.WXK_DELETE or keycode == wx.WXK_BACK:
            #self.DoDelete()

    def ZoomThenRecenter(self, zoomIn=True):
            self.DoScale(zoomIn, bScrollToCenter=True)

    def ToggleGrid(self):
            self.drawGrid = not self.drawGrid
            self.RedrawAssembly()
            
    # Options are:
    #    SIMFW_FULLDISPLAY
    #    SIMFW_GROWTHDISPLAY
    #    SIMFW_NODISPLAY

    def ToggleDisplay(self):
            self.drawUpdates = (self.drawUpdates + 1) % 2
            if self.drawUpdates == SIMFW_FULLDISPLAY:
                if len(self.displayPatterns) == 0:
                    SimUtils.globalApp.OutputMessage("Switched display mode:  Displaying all tiles in the assembly")
                else:
                    SimUtils.globalApp.OutputMessage("Switched display mode:  Displaying all tiles matching the following patterns: " + str(self.displayPatterns))
                self.RedrawAssembly()
                self.Refresh()
                if not self.savedCenteredCoords is None:
                    self.ScrollToCoords(self.savedCenteredCoords)
                self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles))
            #elif self.drawUpdates == SIMFW_GROWTHDISPLAY:
            #    SimUtils.globalApp.OutputMessage("Switched display mode:  Displaying only addition of growth tiles")
            elif self.drawUpdates == SIMFW_NODISPLAY:
                self.savedCenteredCoords = self.GetCenterCoords()
                SimUtils.globalApp.OutputMessage("Switched display mode:  No tile additions are being displayed")

    def SetShowMatching(self):
        patternString = ""
        bDiff = False
        for pattern in self.displayPatterns:
            if len(patternString) > 0:
                patternString = patternString + ","
            patternString = patternString + pattern
        dlg = wx.TextEntryDialog(self, "Enter a comma separated list of strings.  Only tiles whose type\nnames contain at least one of these strings will be displayed.\nTo display all tile types, leave the text box blank.", caption="Set display patterns", value=str(patternString))
        if dlg.ShowModal() == wx.ID_OK:
            patternArray = []
            newPatterns = dlg.GetValue()
            newPatterns = str(newPatterns.strip())
            if len(newPatterns) > 0:
                patternArray = newPatterns.split(",")
            bDiff = (self.displayPatterns != patternArray)
            self.displayPatterns = patternArray
        dlg.Destroy()
        if bDiff == True:
            self.RedrawAssembly(checkAll=True)
            self.Refresh()
            
    def ShowHelp(self):
        info = wx.adv.AboutDialogInfo()
        info.SetName("Generic framework for a self-assembly simulator")
        info.SetVersion("0.1.0")
        helpStr = "\nList of commands:\n\n"
        helpStr = helpStr + "  l : select tile system definition file to load\n"
        helpStr = helpStr + "  s : save current assembly\n"
        helpStr = helpStr + "  r : reset current tile system to seed\n"
        helpStr = helpStr + "  f : simulate one step\n"
        helpStr = helpStr + "  g : begin fast forward simulation\n"
        helpStr = helpStr + "  q : quit fast forward simulation\n"
        helpStr = helpStr + "  p : print list of frontier locations\n"
        helpStr = helpStr + "  t : toggle display of grid lines\n"
        helpStr = helpStr + "  d : toggle display updates of tiles to assembly\n"
        helpStr = helpStr + " +/-: zoom in/out\n"
        helpStr = helpStr + "<left mouse click>:  zoom in on clicked location\n"
        helpStr = helpStr + "<right mouse click>:  zoom out of clicked location\n"
        helpStr = helpStr + "  h : show this help dialog"
        info.SetDescription(helpStr)
        wx.adv.AboutBox(info)

    def SetMouseOverText(self, coords):
        tileTypeName = ""
        tile = self.assembly.GetLoc((coords[0], coords[1]))
        if not tile is None:
            tileTypeName = "Tile type: " + tile.type.name
        else:
            floc = Frontier.FrontierLocation((coords[0], coords[1]),"")
            if floc in self.assembly.frontier:
                tileTypeName = "Frontier location"
                for f in self.assembly.frontier:
                    if f == floc:
                        tileTypeName = "Frontier location, allows " + str(f.fittingTypeNames)
            
        msg = "Coords: (" + str(int(coords[0])) + "," + str(int(coords[1])) + ")"
        self.SendStatusBarUpdate(0, msg)
        self.SendStatusBarUpdate(2, tileTypeName)

    def OnMouse(self, event):
        if self.showMouseOver == True:
            x,y = self.ConvertEventCoords(event)
            x2,y2 = self.CoordsFromPixelPos(x, y)
            self.SetMouseOverText((x2, y2))
            
        if self.mouseMode == SIMFW_MODE_SELECT:
            if event.RightUp():
                self.rubberBand.reset()
            else:
                self.rubberBand.HandleMouseEvents(event)
        else:
            if event.LeftDown():
                if self.mouseMode == SIMFW_MODE_ZOOM:
                    x,y = self.ConvertEventCoords(event)
                    x2,y2 = self.CoordsFromPixelPos(x, y)
                    self.DoScale(True, bScrollToCenter=False, scrollToCoords=(x2,y2))
                    #self.ScrollToCoords((x2,y2))
                    sz = self.GetSize()
                    self.WarpPointer(sz[0]/2, sz[1]/2)
    
            elif event.RightDown():
                if self.mouseMode == SIMFW_MODE_ZOOM:
                    x,y = self.ConvertEventCoords(event)
                    x2,y2 = self.CoordsFromPixelPos(x, y)
                    self.DoScale(False, bScrollToCenter=False, scrollToCoords=(x2,y2))
                    #self.ScrollToCoords((x2,y2))
                    sz = self.GetSize()
                    self.WarpPointer(sz[0]/2, sz[1]/2)

        #elif event.Dragging() or event.LeftUp():


    def ColoredPen(self, color, width = 2):
        if not self.pen_cache.has_key( (color, width) ):
            self.pen_cache[(color, width)] = wx.Pen(color, width)
        return self.pen_cache[(color, width)]

    def ColoredBrush(self, color):
        if not self.brush_cache.has_key(color):
            self.brush_cache[color] = wx.Brush(color)
        return self.brush_cache[color]

    def OnPaint(self, event):
        self._lock.acquire()

        dc = wx.PaintDC(self)

        bg = wx.Brush("WHITE")#(self.GetBackgroundColour())
        dc.SetBackground(bg)
        dc.Clear()

        # use PrepareDC to set position correctly
        self.PrepareDC(dc)
        # we need to clear the dc BEFORE calling PrepareDC
        
        dc.SetUserScale(self.drawScale, self.drawScale)
        
        self.DrawGrid(dc)
        
        for (i,j,loc) in self.drawnAssembly:
            self.DrawAtCoords(dc, i, j, loc)
                
        self._lock.release()
        self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.GetNumTiles()))
        return

    def DrawGrid(self, pdc):
        return

    def ScrollToCoords(self, coords):
        (x,y) = self.GetPixelsForCoords(coords[0], coords[1])
        x = x - self.sideLen/2
        y = y - self.sideLen/2
        x = x * self.drawScale
        y = y * self.drawScale
        (xs, ys) = self.GetScrollPixelsPerUnit()
        (xsz,ysz) = self.GetSize()
        self.Scroll((x-xsz/2+self.sideLen*self.drawScale)/xs, (y-ysz/2+self.sideLen*self.drawScale)/ys)

    def AddToDrawQueue(self, (coords, tileName)):
        if self.NameMatchesPatterns(tileName) == True:
            self._lock.acquire()
            self.DrawQueue.append(coords)
            self.drawnAssembly.append((coords[0], coords[1], self.assembly.GetLoc(coords)))
            self._lock.release()
            self.DrawAssemblyUpdate("new")
        
    def CanvasSizeUpdate(self, centerCoords=None):
        if centerCoords == None:
            centerCoords = self.GetCenterCoords()
        if self.drawUpdates == SIMFW_NODISPLAY:
            return
        wx.CallAfter(pub.getDefaultPublisher().sendMessage, "sizeCanvas", coords = centerCoords)
        
    def DoCanvasSizeUpdate(self, coords=None):
        self.SetVirtualSize((self.width*self.drawScale, self.height*self.drawScale))
        if coords is None:
            coords = self.GetCenterCoords()
        self.ScrollToCoords(coords)
        self.Refresh()

    def DrawAssemblyUpdate(self, msg):
        if self.drawUpdates == SIMFW_NODISPLAY:
            return
        wx.CallAfter(pub.getDefaultPublisher().sendMessage, "drawAssembly", msg = msg)
        
    def DoDrawAssemblyUpdate(self, msg):
        if self.drawUpdates == SIMFW_NODISPLAY:
            return

        if msg == "all":
            self.Refresh()
            return
        elif msg == "allAndCheck":
        #    self.DoInitializeDrawing(dc)
            self.DoRedrawAssembly()

        self._lock.acquire()
               
        dc = wx.ClientDC(self)
            
        self.PrepareDC(dc)
        # we need to clear the dc BEFORE calling PrepareDC
        dc.SetUserScale(self.drawScale, self.drawScale)

        #print("in drawupdate, queue len: " + str(len(self.DrawQueue)))
        if (msg == "new"):
            for e in self.DrawQueue:
                self.DrawAtCoords(dc, e[0], e[1], self.assembly.GetLoc(e))
            self.DrawQueue = []        
        self._lock.release()

        #dc.BeginDrawing()
        #  Iterate the grid and draw tile types
        #for i in range(self.width):
        #    for j in range(self.height):
        #        if self.assembly.grid[i][j] is not None:
        #            self.DrawAtCoords(dc, i, j, self.assembly.grid[i][j])

        #for e in self.RemoveQueue:
        #    dc.RemoveId(e)
            #pen = self.ColoredPen("BLACK")
            #dc.SetPen(pen)
            #dc.SetBrush(self.ColoredBrush("WHITE"))
            #dc.DrawRectangle(0,0,300,300)                         
        #self.RemoveQueue = []
        #dc.EndDrawing()
        
    def DoRedrawAssembly(self):
        self._lock.acquire()
        self.drawnAssembly = []
        for i in range(self.assembly.minX, self.assembly.maxX+1):
            for j in range(self.assembly.minY, self.assembly.maxY+1):
                loc = self.assembly.GetLoc((i,j))
                if loc is not None:
                    if self.NameMatchesPatterns(loc.type.name) == True:
                        self.drawnAssembly.append((i,j,loc))
        self._lock.release()
        self.Refresh()
        
    def InitializeDrawing(self):
        self.DrawAssemblyUpdate("all")
        
#===============================================================================
#     def DoInitializeDrawing(self, dc):
#         #dc.BeginDrawing()
#         
#         bg = wx.Brush("WHITE")#(self.GetBackgroundColour())
#         dc.SetBackground(bg)
#         dc.Clear()
# 
#         self.numVert = self.width / self.sideLen
#         self.numHoriz = self.height / self.sideLen
# 
#         if self.drawGrid == True:
#             dc.SetPen(self.gridPen)
#             for i in range(0, self.numVert+1):
#                 x = i * self.sideLen
#                 dc.DrawLine(x, self.height, x, self.height-self.numVert*self.sideLen)
#             for i in range(0, self.numHoriz+1):
#                 y = i * self.sideLen
#                 dc.DrawLine(0, self.height-y, self.numHoriz*self.sideLen, self.height-y)
#          
#         #dc.EndDrawing()
#         #self.Refresh()
#===============================================================================
        
    def RedrawAssembly(self, checkAll=False):
        if checkAll == True:
            self.DrawAssemblyUpdate("allAndCheck")
        else:
            self.DrawAssemblyUpdate("all")

    def DoScale(self, bIncrease, bScrollToCenter=False, scrollToCoords=None):
        if bScrollToCenter == True:
            coords = self.GetCenterCoords()
        else:
            coords = scrollToCoords
        if bIncrease == True:
            self.drawScale = self.drawScale * 2.0
        else:
            self.drawScale = self.drawScale / 2.0
        print("Scale set to " + str(self.drawScale))
        self.CanvasSizeUpdate(coords)

    def GetPixelsForCoords(self, xCoord, yCoord):
        xCoord = xCoord + int(self.numVert/2)
        yCoord = yCoord + int(self.numHoriz/2)
        xPixel = (xCoord * self.sideLen)# * self.drawScale
        yPixel = (self.height - ((yCoord+1) * self.sideLen))# * self.drawScale
        return (xPixel,yPixel)
  
    def CoordsFromPixelPos(self, x, y, bClosest=True):
        # bClosest isn't used in this default implementation but may be useful in grids where some pixels aren't on valid coordinates
        x1 = int(x / (self.sideLen*self.drawScale))
        y1 = int(((self.height*self.drawScale)-y) / (self.sideLen*self.drawScale))

        x1 = x1 - int(self.numVert/2)
        y1 = y1 - int(self.numVert/2)

        return (int(x1), int(y1))
    
    def NameMatchesPatterns(self, name):
        if len(self.displayPatterns) == 0:
            return True
        else:
            for pattern in self.displayPatterns:
                if pattern in name:
                    return True
            return False
        
    def DrawAtCoords(self, dc, xCoord, yCoord, onode):
        if onode is None:
            return

        #MJPTODO: only need to set brush once for the dc
        #dc.SetBrush(self.ColoredBrush("WHITE"))

        #Doing this elsewhere now:
        #if self.NameMatchesPatterns(onode.type.name) == True:
        (xPixel,yPixel) = self.GetPixelsForCoords(xCoord, yCoord)
        # Adjust coords for correct offset of tile images
        xPixel = xPixel - IMG_OFFSET_HORIZ
        yPixel = yPixel - IMG_OFFSET_VERT
        #print("drawing tile for coords " + str((xCoord,yCoord)) + " at pixel " + str((xPixel,yPixel)))
        onode.type.DrawAtLocation(xPixel, yPixel, dc)


        #else:
        #    dc.DrawRectangle(xPixel, yPixel, 96, 96)

