#!/usr/bin/env python
#----------------------------------------------------------------------------
# Name:         setup.py
# Purpose:      A script for packaging OritatamiSim as a Windows executable
#
# Author:       Matt Patitz
#
# Created:      
# RCS-ID:       
# Copyright:    
# License:      
#----------------------------------------------------------------------------

"""
This program packages the components of OritatamiSim into a Windows executable package

Requirements:
    This application requires Python 2.x

Dependencies:
    py2exe    (install using 'pip install http://sourceforge.net/projects/py2exe/files/latest/download?source=files')
"""


from distutils.core import setup
import py2exe
import os
import glob


def find_data_files(source,target,patterns):
    """Locates the specified data-files and returns the matches
    in a data_files compatible format.

    source is the root of the source data tree.
        Use '' or '.' for current directory.
    target is the root of the target data tree.
        Use '' or '.' for the distribution directory.
    patterns is a sequence of glob-patterns for the
        files you want to copy.
    """
    if glob.has_magic(source) or glob.has_magic(target):
        raise ValueError("Magic not allowed in src, target")
    ret = {}
    for pattern in patterns:
        pattern = os.path.join(source,pattern)
        for filename in glob.glob(pattern):
            if os.path.isfile(filename):
                targetpath = os.path.join(target,os.path.relpath(filename,source))
                path = os.path.dirname(targetpath)
                ret.setdefault(path,[]).append(filename)
    return sorted(ret.items())

setup(
    name = 'Oritatami Simulator',
    description = 'Oritatami simulator and system development environment',
    version = '1.0',

    data_files=find_data_files('','',[
        'samples/*',
        'images/*',
        'docs/*'
    ]),

    windows = [
                  {
                      'script': 'OritatamiSim.py',
                  }
              ],

    options = {
                  'py2exe': {
					  'dist_dir': 'packaging/OritatamiSim-win-exe',
                      'unbuffered': True,
					  #'bundle_files': 2,
                      'optimize': 2,
                      'includes': ['wx', 'os', 'random', 'math', 're', 'sys', 'thread', 'itertools', 'time', 'threading'],
                  }
              },
) 
